@@ -137,7 +137,7 @@ def _register_timeline(optimizer, models, parameter_names, parent_name=None):
 
 
 class _DistributedOptimizer(torch.optim.Optimizer):
-    def __init__(self, params, model):
+    def __init__(self, params, model, backward_passes_per_step=1):
         super(self.__class__, self).__init__(params)
 
         named_parameters, models = _check_named_parameters(self, model)
@@ -150,6 +150,9 @@ class _DistributedOptimizer(torch.optim.Optimizer):
         self._should_synchronize = True
         self._timeline_hook_handles = []
         self._use_timeline = False
+        self._backward_passes_per_step = backward_passes_per_step
+        self._allreduce_delay = {v: self._backward_passes_per_step
+                                 for _, v in sorted(named_parameters)}
         if os.getenv('BLUEFOG_TIMELINE'):
             self.turn_on_timeline()
         if bf.size() > 1:
@@ -169,8 +172,16 @@ class _DistributedOptimizer(torch.optim.Optimizer):
     def _make_hook(self, p):
         def hook(*ignore):
             assert not p.grad.requires_grad
-            handle = self._allreduce_grad_async(p)
-            self._handles[p] = handle
+            if self._allreduce_delay[p] <= 0:
+                raise AssertionError(
+                    "Unexpected behavior: backward computation were computed "
+                    "more than num_steps_per_communication times before call "
+                    "to step(). Adjust num_steps_per_communication to "
+                    "accumulate gradients locally.")
+            self._allreduce_delay[p] -= 1
+            if self._allreduce_delay[p] == 0:
+                handle = self._allreduce_grad_async(p)
+                self._handles[p] = handle
 
         return hook
 
@@ -203,8 +214,10 @@ class _DistributedOptimizer(torch.optim.Optimizer):
             if handle is None:
                 handle = self._allreduce_grad_async(p)
                 self._handles[p] = handle
+
         for p, handle in self._handles.items():
             output = bf.synchronize(handle)
+            self._allreduce_delay[p] = self._backward_passes_per_step
             p.grad.set_(output)
         self._handles.clear()
 
@@ -254,10 +267,13 @@ class _DistributedOptimizer(torch.optim.Optimizer):
             )
         return super(self.__class__, self).zero_grad()
 
-
-class _DistributedNeighborAllreduceOptimizer(torch.optim.Optimizer):
+class _DistributedReduceOptimizer(torch.optim.Optimizer):
     """ A distributed optimizer wrapper over torch optimizer.
 
+    Arguments:
+        reduce_type: either to be "allreduce" or "neighbor_allreduce" to decide the reducing
+                     method for communication.
+
     Note: Unlike the _DistributedOptimizer class that registers hook for each named parameters,
     triggers the allreduce_nonblocking after gradient computation is finished, and updates the
     parameters at last step function. We will trigger the win_put ops that puts the weights
@@ -275,12 +291,18 @@ class _DistributedNeighborAllreduceOptimizer(torch.optim.Optimizer):
         w_{i+1, k} = Neighbor_Average( w_{i, k} - lr * local_grad(w_{i, k}) )
     """
 
-    def __init__(self, params, model):
+    def __init__(self, params, model, reduce_type, num_steps_per_communication=1):
         super(self.__class__, self).__init__(params)
 
         named_parameters, models = _check_named_parameters(self, model)
-        self.neighbor_weights = None
+        # knobs for neighbor communication behavior
         self.self_weight = None
+        self.neighbor_weights = None
+        self.send_neighbors = None
+        self.neighbor_machine_weights = None
+        self.send_neighbor_machines = None
+        self.enable_topo_check = False
+
         self._models = models
         self._parameter_names = {v: k for k, v in sorted(named_parameters)}
         self._handles = {}
@@ -289,6 +311,20 @@ class _DistributedNeighborAllreduceOptimizer(torch.optim.Optimizer):
         self._should_synchronize = True
         self._timeline_hook_handles = []
         self._use_timeline = False
+        self._num_steps_per_communication = num_steps_per_communication
+        self._reduce_type_str = reduce_type
+        # _reduce_method: 0 for allreduce, and 1 for neighbor_allreduce
+        if self._reduce_type_str == "allreduce":
+            self._reduce_method = 0
+        elif self._reduce_type_str == "neighbor.allreduce":
+            self._reduce_method = 1
+        elif self._reduce_type_str == "hierarchical.neighbor.allreduce":
+            self._reduce_method = 2
+        else:
+            raise ValueError("Unknown reduce type for internal class _DistributedReduceOptimizer")
+
+        self._reduce_delay = {v: self._num_steps_per_communication
+                              for _, v in sorted(named_parameters)}
         if os.getenv('BLUEFOG_TIMELINE'):
             self.turn_on_timeline()
         if bf.size() > 1:
@@ -310,19 +346,53 @@ class _DistributedNeighborAllreduceOptimizer(torch.optim.Optimizer):
                     # End forward computation timeline
                     bf.timeline_end_activity(parent_name+'.'+name)
                 if p.requires_grad:
-                    handle = self._neighbor_allreduce_data_async(p)
-                    self._handles[p] = handle
+                    if self._reduce_delay[p] <= 0:
+                        raise AssertionError(
+                            "Unexpected behavior: forward computation were computed "
+                            "more than num_steps_per_communication times before call "
+                            "to step(). Adjust num_steps_per_communication to "
+                            "accumulate gradients locally.")
+                    self._reduce_delay[p] -= 1
+                    if self._reduce_delay[p] == 0:
+                        if self._reduce_method == 0:
+                            handle = self._allreduce_data_async(p)
+                        elif self._reduce_method == 1:
+                            handle = self._neighbor_allreduce_data_async(p)
+                        elif self._reduce_method == 2:
+                            handle = self._hierarchical_neighbor_allreduce_data_async(p)
+                        elif self._reduce_method == -1:
+                            handle = None
+                        else:
+                            raise ValueError(
+                                "Unknown reduce method. Do not change _reduce_method manually.")
+                        self._handles[p] = handle
         return hook
 
     def _neighbor_allreduce_data_async(self, p):
         name = self._parameter_names.get(p)
         handle = bf.neighbor_allreduce_nonblocking(p.data, name=name, self_weight=self.self_weight,
-                                                   neighbor_weights=self.neighbor_weights)
+                                                   neighbor_weights=self.neighbor_weights,
+                                                   send_neighbors=self.send_neighbors,
+                                                   enable_topo_check=self.enable_topo_check)
+        return handle
+
+    def _hierarchical_neighbor_allreduce_data_async(self, p):
+        name = self._parameter_names.get(p)
+        handle = bf.hierarchical_neighbor_allreduce_nonblocking(
+            p.data, name=name, self_weight=self.self_weight,
+            neighbor_machine_weights=self.neighbor_machine_weights,
+            send_neighbor_machines=self.send_neighbor_machines,
+            enable_topo_check=self.enable_topo_check)
+        return handle
+
+    def _allreduce_data_async(self, p):
+        name = self._parameter_names.get(p)
+        handle = bf.allreduce_nonblocking(p.data, average=True, name=name)
         return handle
 
     def turn_on_timeline(self):
         handles = _register_timeline(
-            self, self._models, self._parameter_names, 'neighbor.allreduce')
+            self, self._models, self._parameter_names, self._reduce_type_str)
         self._timeline_hook_handles.extend(handles)
         self._use_timeline = True
 
@@ -332,21 +402,25 @@ class _DistributedNeighborAllreduceOptimizer(torch.optim.Optimizer):
         self._timeline_hook_handles.clear()
         self._use_timeline = False
 
+    def use_allreduce_in_communication(self):
+        self._reduce_method = 0
+
+    def use_neighbor_allreduce_in_communication(self):
+        self._reduce_method = 1
+
+    def use_hierarchical_neighbor_allreduce_in_communication(self):
+        self._reduce_method = 2
+
+    def use_empty_function_in_communication(self):
+        self._reduce_method = -1
+
     def synchronize(self):
-        missing_p = self._requires_update - set(self._handles.keys())
-        for p in missing_p:
-            handle = self._neighbor_allreduce_data_async(p)
-            self._handles[p] = handle
-
-        for p, handle in self._handles.items():
-            if handle is None:
-                handle = self._neighbor_allreduce_data_async(p)
-                self._handles[p] = handle
-
         with torch.no_grad():
             for p, handle in self._handles.items():
-                output = bf.synchronize(handle)
-                p.set_(output)
+                if handle is not None:
+                    output = bf.synchronize(handle)
+                    p.set_(output)
+                self._reduce_delay[p] = self._num_steps_per_communication
         self._handles.clear()
 
         self._synchronized = True
@@ -388,9 +462,9 @@ class _DistributedNeighborAllreduceOptimizer(torch.optim.Optimizer):
         return super(self.__class__, self).step(closure)
 
 
-class _DistributedBluefogOptimizer(torch.optim.Optimizer):
+class _DistributedWinOptimizer(torch.optim.Optimizer):
 
-    def __init__(self, params, model, pull_style):
+    def __init__(self, params, model, num_steps_per_communication, pull_style):
         super(self.__class__, self).__init__(params)
 
         if pull_style:
@@ -407,6 +481,9 @@ class _DistributedBluefogOptimizer(torch.optim.Optimizer):
         self._synchronized = False
         self._should_synchronize = True
         self._use_timeline = False
+        self._num_steps_per_communication = num_steps_per_communication
+        self._bluefog_delay = {v: self._num_steps_per_communication
+                               for _, v in sorted(named_parameters)}
         self._timeline_hook_handles = []
         if os.getenv('BLUEFOG_TIMELINE'):
             self.turn_on_timeline()
@@ -432,10 +509,18 @@ class _DistributedBluefogOptimizer(torch.optim.Optimizer):
                 if not module.training:
                     continue
                 if p.requires_grad:
-                    handle = bf.win_put_nonblocking(
-                        tensor=p.data, name=parent_name+'.'+name,
-                        dst_weights=self.dst_weights, require_mutex=False)
-                    self._handles[p] = handle
+                    if self._bluefog_delay[p] <= 0:
+                        raise AssertionError(
+                            "Unexpected behavior: forward computation were computed "
+                            "more than num_steps_per_communication times before call "
+                            "to step(). Adjust num_steps_per_communication to "
+                            "accumulate gradients locally.")
+                    self._bluefog_delay[p] -= 1
+                    if self._bluefog_delay[p] == 0:
+                        handle = bf.win_put_nonblocking(
+                            tensor=p.data, name=parent_name+'.'+name,
+                            dst_weights=self.dst_weights, require_mutex=False)
+                        self._handles[p] = handle
         return hook
 
     def _make_get_hook(self, parent_name):
@@ -447,10 +532,18 @@ class _DistributedBluefogOptimizer(torch.optim.Optimizer):
                 if not module.training:
                     continue
                 if p.requires_grad:
-                    handle = bf.win_get_nonblocking(
-                        name=parent_name+'.'+name, src_weights=self.src_weights,
-                        require_mutex=True)
-                    self._handles[p] = handle
+                    if self._bluefog_delay[p] <= 0:
+                        raise AssertionError(
+                            "Unexpected behavior: forward computation were computed "
+                            "more than num_steps_per_communication times before call "
+                            "to step(). Adjust num_steps_per_communication to "
+                            "accumulate gradients locally.")
+                    self._bluefog_delay[p] -= 1
+                    if self._bluefog_delay[p] == 0:
+                        handle = bf.win_get_nonblocking(
+                            name=parent_name+'.'+name, src_weights=self.src_weights,
+                            require_mutex=True)
+                        self._handles[p] = handle
         return hook
 
     def _register_window(self):
@@ -502,6 +595,7 @@ class _DistributedBluefogOptimizer(torch.optim.Optimizer):
             for p, handle in self._handles.items():
                 _ = bf.win_wait(handle)
                 name = self._parameter_names.get(p)
+                self._bluefog_delay[p] = self._num_steps_per_communication
                 # Update p to the average of neighbors.
                 p.set_(bf.win_update(name=name, require_mutex=True))
 
@@ -529,7 +623,7 @@ class _DistributedBluefogOptimizer(torch.optim.Optimizer):
 
 class _DistributedPushSumOptimizer(torch.optim.Optimizer):
 
-    def __init__(self, params, model):
+    def __init__(self, params, model, num_steps_per_communication):
         super(self.__class__, self).__init__(params)
 
         # use to control the behavior of win_accumulate dynamically.
@@ -548,6 +642,9 @@ class _DistributedPushSumOptimizer(torch.optim.Optimizer):
         self._synchronized = False
         self._should_synchronize = True
         self._use_timeline = False
+        self._num_steps_per_communication = num_steps_per_communication
+        self._pushsum_delay = {v: self._num_steps_per_communication
+                               for _, v in sorted(named_parameters)}
         self._timeline_hook_handles = []
         if bf.size() > 1:
             self._register_window()
@@ -587,15 +684,22 @@ class _DistributedPushSumOptimizer(torch.optim.Optimizer):
                 if not module.training:
                     continue
                 if p.requires_grad:
-                    ps_weights = self._named_ps_weights[full_name]
-                    extended_parameter = torch.cat((p.data.view(-1), ps_weights), 0)
-                    self._named_extension_parameters[name] = extended_parameter
-                    handle = bf.win_accumulate_nonblocking(
-                        tensor=extended_parameter, name=full_name,
-                        dst_weights=self.dst_weights,
-                        require_mutex=True)
-
-                    self._handles[p] = handle
+                    if self._pushsum_delay[p] <= 0:
+                        raise AssertionError(
+                            "Unexpected behavior: forward computation were computed "
+                            "more than num_steps_per_communication times before call "
+                            "to step(). Adjust num_steps_per_communication to "
+                            "accumulate gradients locally.")
+                    self._pushsum_delay[p] -= 1
+                    if self._pushsum_delay[p] == 0:
+                        ps_weights = self._named_ps_weights[full_name]
+                        extended_parameter = torch.cat((p.data.view(-1), ps_weights), 0)
+                        self._named_extension_parameters[name] = extended_parameter
+                        handle = bf.win_accumulate_nonblocking(
+                            tensor=extended_parameter, name=full_name,
+                            dst_weights=self.dst_weights,
+                            require_mutex=True)
+                        self._handles[p] = handle
         return hook
 
     def turn_on_timeline(self):
@@ -636,6 +740,7 @@ class _DistributedPushSumOptimizer(torch.optim.Optimizer):
             for p, handle in self._handles.items():
                 _ = bf.win_wait(handle)
                 name = self._parameter_names.get(p)
+                self._pushsum_delay[p] = self._num_steps_per_communication
                 extended_parameter = self._named_extension_parameters[name]
                 extended_parameter.mul_(self.self_weight)
                 # Last dimension is the push_sum weights and we want parameter / weight
@@ -668,7 +773,8 @@ class _DistributedPushSumOptimizer(torch.optim.Optimizer):
         return super(self.__class__, self).step(closure)
 
 
-def DistributedPushSumOptimizer(optimizer, model):
+def DistributedPushSumOptimizer(optimizer, model,
+                                num_steps_per_communication=1):
     """
     An distributed optimizer that wraps another torch.optim.Optimizer through
     win_accumulate ops to implement the gradient push algorithm.
@@ -680,15 +786,40 @@ def DistributedPushSumOptimizer(optimizer, model):
     Arguments:
         optimizer: Optimizer to use for computing gradients and applying updates.
         model: The model or a list of models you want to train with.
+        num_steps_per_communication: Number of expected model forward function calls before each
+                                     communication. This allows local model parameter updates
+                                     per num_steps_per_communication before reducing them over
+                                     distributed computation resources.
+
+    Example for two scenarios to use num_steps_per_communication:
+        Scenario 1) Local accumulation of gradient without update model.
+                    (Used in large batch size or large model cases)
+        >>> opt = bf.DistributedPushSumOptimizer(optimizer, model, num_steps_per_communication=J)
+        >>> opt.zero_grad()
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     loss.backward()
+        >>> opt.step()  # PushSum reducing happens here
+        Scenario 2) Local updating the model. (Used in case that decreasing the communication).
+        >>> opt = bf.DistributedPushSumOptimizer(optimizer, model, num_steps_per_communication=J)
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     opt.zero_grad()
+        >>>     loss.backward()
+        >>>     opt.step()  # PushSum reducing happens at the last iteration
     """
     cls = type(
         optimizer.__class__.__name__,
         (optimizer.__class__,),
         dict(_DistributedPushSumOptimizer.__dict__),
     )
-    return cls(optimizer.param_groups, model)
+    return cls(optimizer.param_groups, model, num_steps_per_communication)
 
-def DistributedPullGetOptimizer(optimizer, model):
+
+def DistributedPullGetOptimizer(optimizer, model,
+                                num_steps_per_communication=1):
     """
     An distributed optimizer that wraps another torch.optim.Optimizer with
     pull model average through bf.win_get ops.
@@ -696,6 +827,29 @@ def DistributedPullGetOptimizer(optimizer, model):
     Arguments:
         optimizer: Optimizer to use for computing gradients and applying updates.
         model: The model or a list of models you want to train with.
+        num_steps_per_communication: Number of expected model forward function calls before each
+                                     communication. This allows local model parameter updates
+                                     per num_steps_per_communication before reducing them over
+                                     distributed computation resources.
+
+    Example for two scenarios to use num_steps_per_communication:
+        Scenario 1) Local accumulation of gradient without update model.
+                    (Used in large batch size or large model cases)
+        >>> opt = bf.DistributedPullGetOptimizer(optimizer, model, num_steps_per_communication=J)
+        >>> opt.zero_grad()
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     loss.backward()
+        >>> opt.step()  # PullGet communication happens here
+        Scenario 2) Local updating the model. (Used in case that decreasing the communication).
+        >>> opt = bf.DistributedPullGetOptimizer(optimizer, model, num_steps_per_communication=J)
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     opt.zero_grad()
+        >>>     loss.backward()
+        >>>     opt.step()  # PullGet communication happens at the last iteration
 
     Returned optimizer has two extra parameters `src_weights` and `force_barrier`.
     Set src_weights dictionary as {rank: scaling} differently per iteration to achieve
@@ -705,71 +859,209 @@ def DistributedPullGetOptimizer(optimizer, model):
     cls = type(
         optimizer.__class__.__name__,
         (optimizer.__class__,),
-        dict(_DistributedBluefogOptimizer.__dict__),
+        dict(_DistributedWinOptimizer.__dict__),
     )
-    return cls(optimizer.param_groups, model, pull_style=True)
+    return cls(optimizer.param_groups, model, num_steps_per_communication, pull_style=True)
 
 
-def DistributedBluefogOptimizer(optimizer, model):
+def DistributedWinPutOptimizer(optimizer, model,
+                               num_steps_per_communication=1):
     """An distributed optimizer that wraps another torch.optim.Optimizer with
     pull model average through bf.win_put ops.
 
     Arguments:
         optimizer: Optimizer to use for computing gradients and applying updates.
         model: The model or a list of models you want to train with.
+        num_steps_per_communication: Number of expected model forward function calls before each
+                                     communication. This allows local model parameter updates
+                                     per num_steps_per_communication before reducing them over
+                                     distributed computation resources.
 
     Returned optimizer has two extra parameters `dst_weights` and `force_barrier`.
     Set dst_weights dictionary as {rank: scaling} differently per iteration to achieve
     win_put over dynamic graph behavior. If force_barrier is True, a barrier function
     will put at `step()` to synchronous processes.
-
-    Example:
-        >>> import bluefog.torch as bf
-        >>> ...
-        >>> bf.init()
-        >>> optimizer = optim.SGD(model.parameters(), lr=lr * bf.size())
-        >>> optimizer = bf.DistributedBluefogOptimizer(optimizer, model)
     """
     # We dynamically create a new class that inherits from the optimizer that was passed in.
     # The goal is to override the `step()` method.
     cls = type(
         optimizer.__class__.__name__,
         (optimizer.__class__,),
-        dict(_DistributedBluefogOptimizer.__dict__),
+        dict(_DistributedWinOptimizer.__dict__),
     )
-    return cls(optimizer.param_groups, model, pull_style=False)
+    return cls(optimizer.param_groups, model, num_steps_per_communication, pull_style=False)
 
 
-def DistributedNeighborAllreduceOptimizer(optimizer, model):
+def DistributedAllreduceOptimizer(optimizer, model,
+                                  num_steps_per_communication=1):
     """
-    An distributed optimizer that wraps another torch.optim.Optimizer through
-    neighbor_allreduce ops.
-
-    Returned optimizer has two extra parameters `self_weight` and `neighbor_weights`.
-    Set self_weight as some scalar and dst_weights dictionary as {rank: scaling} differently
-    per iteration to achieve win_put over dynamic graph behavior.
+    An distributed optimizer that wraps another torch.optim.Optimizer through allreduce ops.
+    The communication for allreduce is applied on the parameters when forward propagation happens.
 
     Arguments:
         optimizer: Optimizer to use for computing gradients and applying updates.
         model: The model or a list of models you want to train with.
+        num_steps_per_communication: Number of expected model forward function calls before each
+                                     communication. This allows local model parameter updates
+                                     per num_steps_per_communication before reducing them over
+                                     distributed computation resources.
+
+    Example for two scenarios to use num_steps_per_communication.
+        Scenario 1) Local accumulation of gradient without update model.
+                    (Used in large batch size or large model cases)
+
+        >>> opt = bf.DistributedAllreduceOptimizer(optimizer, model,
+                                                   num_steps_per_communication=J)
+        >>> opt.zero_grad()
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     loss.backward()
+        >>> opt.step()  # Allreducing happens here
+
+        Scenario 2) Local updating the model. (Used in case that decreasing the communication).
+
+        >>> opt = bf.DistributedAllreduceOptimizer(optimizer, model,
+                                                   num_steps_per_communication=J)
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     opt.zero_grad()
+        >>>     loss.backward()
+        >>>     opt.step()  # Allreducing happens at the last iteration
+    """
+    # We dynamically create a new class that inherits from the optimizer that was passed in.
+    # The goal is to override the `step()` method with allreduce implementation.
+    cls = type(
+        optimizer.__class__.__name__,
+        (optimizer.__class__,),
+        dict(_DistributedReduceOptimizer.__dict__),
+    )
+    return cls(optimizer.param_groups, model, "allreduce", num_steps_per_communication)
+
+
+def DistributedNeighborAllreduceOptimizer(optimizer, model,
+                                          num_steps_per_communication=1):
+    """
+    An distributed optimizer that wraps another torch.optim.Optimizer through
+    neighbor_allreduce ops over parameters.
+
+    Returned optimizer has three extra parameters `self_weight`, `neighbor_weights` and
+    `send_neighbors` to control the behavior of neighbor allreduce. Changing the values
+    of these knobs to achieve dynamic topologies.
+
+    Arguments:
+        optimizer: Optimizer to use for computing gradients and applying updates.
+        model: The model or a list of models you want to train with.
+        num_steps_per_communication: Number of expected model forward function calls before each
+                                     communication. This allows local model parameter updates
+                                     per num_steps_per_communication before reducing them over
+                                     distributed computation resources.
     """
     # We dynamically create a new class that inherits from the optimizer that was passed in.
     # The goal is to override the `step()` method with neighbor_allreduce implementation.
     cls = type(
         optimizer.__class__.__name__,
         (optimizer.__class__,),
-        dict(_DistributedNeighborAllreduceOptimizer.__dict__),
+        dict(_DistributedReduceOptimizer.__dict__),
     )
-    return cls(optimizer.param_groups, model)
+    return cls(optimizer.param_groups, model, "neighbor.allreduce", num_steps_per_communication)
 
 
-def DistributedAllreduceOptimizer(optimizer, model):
+def DistributedHierarchicalNeighborAllreduceOptimizer(optimizer, model,
+                                                      num_steps_per_communication=1):
     """
-    An distributed optimizer that wraps another torch.optim.Optimizer through allreduce ops.
+    An distributed optimizer that wraps another torch.optim.Optimizer through
+    hierarchical_neighbor_allreduce ops over parameters.
+
+    Returned optimizer has three extra parameters `self_weight`, `neighbor_machine_weights` and
+    `send_neighbor_machines` to control the behavior of hierarchical neighbor allreduce. Changing
+    the values of these knobs to achieve dynamic topologies.
 
     Arguments:
         optimizer: Optimizer to use for computing gradients and applying updates.
         model: The model or a list of models you want to train with.
+        num_steps_per_communication: Number of expected model forward function calls before each
+                                     communication. This allows local model parameter updates
+                                     per num_steps_per_communication before reducing them over
+                                     distributed computation resources.
+
+    Warning:
+        The processes within the same machine should provide the same `neighbor_machine_weights` and
+        `send_neighbor_machines` to avoid unexpected behavior.
+
+    Example for two scenarios to use num_steps_per_communication:
+        Scenario 1) Local accumulation of gradient without update model.
+                    (Used in large batch size or large model cases)
+
+        >>> opt = bf.DistributedHierarchicalNeighborAllreduceOptimizer(
+                        optimizer, model, num_steps_per_communication=J)
+        >>> opt.zero_grad()
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     loss.backward()
+        >>> opt.step()  # Neighbor allreducing happens here
+
+        Scenario 2) Local updating the model. (Used in case that decreasing the communication).
+
+        >>> opt = bf.DistributedHierarchicalNeighborAllreduceOptimizer(
+                        optimizer, model, num_steps_per_communication=J)
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     opt.zero_grad()
+        >>>     loss.backward()
+        >>>     opt.step()  # Neighbor allreducing happens at the last iteration
+    """
+    cls = type(
+        optimizer.__class__.__name__,
+        (optimizer.__class__,),
+        dict(_DistributedReduceOptimizer.__dict__),
+    )
+    return cls(optimizer.param_groups, model, "hierarchical.neighbor.allreduce",
+               num_steps_per_communication)
+
+
+def DistributedGradientAllreduceOptimizer(optimizer, model,
+                                          backward_passes_per_step=1):
+    """
+    An distributed optimizer that wraps another torch.optim.Optimizer through allreduce ops.
+    The communication happens when backward propagation happens, which is the same as Horovod.
+    In addition, allreduce is applied on gradient instead of parameters.
+
+    Arguments:
+        optimizer: Optimizer to use for computing gradients and applying updates.
+        model: The model or a list of models you want to train with.
+        num_steps_per_communication: Number of expected backward function calls before each
+                                     communication. This allows local model parameter updates
+                                     per num_steps_per_communication before reducing them over
+                                     distributed computation resources.
+
+    Example for two scenarios to use num_steps_per_communication:
+
+        Scenario 1) Local accumulation of gradient without update model.
+                    (Used in large batch size or large model cases)
+
+        >>> opt = bf.DistributedGradientAllreduceOptimizer(optimizer, model,
+        >>>                                                num_steps_per_communication=J)
+        >>> opt.zero_grad()
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     loss.backward()
+        >>> opt.step()  # Allreducing happens here
+
+        Scenario 2) Local updating the model. (Used in case that decreasing the communication).
+
+        >>> opt = bf.DistributedGradientAllreduceOptimizer(optimizer, model,
+        >>>                                                num_steps_per_communication=J)
+        >>> for j in range(J):
+        >>>     output = model(data_batch_i)
+        >>>     loss = ...
+        >>>     opt.zero_grad()
+        >>>     loss.backward()
+        >>>     opt.step()  # Allreducing happens at the last iteration
     """
     # We dynamically create a new class that inherits from the optimizer that was passed in.
     # The goal is to override the `step()` method with an allreduce implementation.
@@ -778,4 +1070,4 @@ def DistributedAllreduceOptimizer(optimizer, model):
         (optimizer.__class__,),
         dict(_DistributedOptimizer.__dict__),
     )
-    return cls(optimizer.param_groups, model)
+    return cls(optimizer.param_groups, model, backward_passes_per_step)
